#!/bin/bash
##
# Licensed to Cloudera, Inc. under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  Cloudera, Inc. licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

. $(cd $(dirname $0) && pwd)/common.sh

log "Starting Livy server (CDP $CDH_VERSION)"

set -ex
export CDP_ROOT=$(cd $(m_readlink "$CDH_HADOOP_HOME/../..") && pwd)

SPARK_CONF_DIR="$CONF_DIR/spark3-conf"
HIVE_CONF_DIR="$CONF_DIR/hive-conf"

gen_livy_conf() {
  log "Generating Livy server configuration..."

  # Because of OPSAPS-25695, the Spark config provided to this script is not complete, so
  # we have to manually invoke the Spark CSD code (repackaged into this CSD) to generate
  # the proper config.
  rm -rf "$CONF_DIR/spark3-conf/yarn-conf"
  deploy_client_config

  local LIVY_CONF="$CONF_DIR/livy3-conf/livy.conf"
  local HADOOP_CONF_DIR="$SPARK_CONF_DIR/yarn-conf"

  # Livy doesn't have a setting to enable or disable TLS, and instead uses the keystore config
  # to signal TLS should be enabled. So remove it from the config file when it's not enabled.
  local SSL_ENABLED=$(read_property "livy.tls.enabled" "$LIVY_CONF")
  if [ "$SSL_ENABLED" = "true" ]; then
    local CREDS_KEY="hadoop.security.credential.provider.path"
    local CREDS=$(read_property "$CREDS_KEY" "$LIVY_CONF")
    if [ -z "$CREDS" ]; then
      log "Credential provider location not set!"
      exit 1
    fi
    replace_spark_conf "livy.keystore.password" "" "$LIVY_CONF"
    replace_spark_conf "livy.key-password" "" "$LIVY_CONF"
    replace_spark_conf "$CREDS_KEY" "" "$LIVY_CONF"
    echo "livy.$CREDS_KEY=$CREDS" >> "$LIVY_CONF"
    if [ "$ENABLE_THRIFT" = "true" ]; then
      echo "livy.server.thrift.use.SSL=true" >> "$LIVY_CONF"
    fi

    # Set FIPS related configs
    FIPS_ENABLED_SYSTEM_PROPERTY='com.safelogic.cryptocomply.fips.approved_only=true'
    if [[ "${CSD_JAVA_OPTS}" == *"${FIPS_ENABLED_SYSTEM_PROPERTY}"* ]]; then
      # By default, Jetty's SSL cipher blacklist is defined using the following regular expression: '^.*_(MD5|SHA|SHA1)$, ^TLS_RSA_.*$, ^SSL_.*$, ^.*_NULL_.*$, ^.*_anon_.*$'
      # To support the new bctls-safelogic.jar, we need to allow 'SHA' (therefore we remove that from the above regex)
      replace_spark_conf "livy.ssl.ui.excludedAlgorithms" "^.*_(MD5|SHA1)$, ^TLS_RSA_.*$, ^SSL_.*$, ^.*_NULL_.*$, ^.*_anon_.*$" "$LIVY_CONF"

      # We want to allow TLSv1.2 only
      replace_spark_conf "livy.ssl.ui.includedProtocols" "TLSv1.2" "$LIVY_CONF"
    fi
  else
    replace_spark_conf "livy.keystore" "" "$LIVY_CONF"
  fi

  if [ -n "$LIVY_PRINCIPAL" ]; then
    echo "livy.server.launch.kerberos.keytab=$CONF_DIR/livy_for_spark3.keytab" >> "$LIVY_CONF"
    echo "livy.server.launch.kerberos.principal=$LIVY_PRINCIPAL" >> "$LIVY_CONF"

    if [ "$ENABLE_SPNEGO" = "true" ]; then
      echo "livy.server.auth.type=kerberos" >> "$LIVY_CONF"
      echo "livy.server.auth.kerberos.keytab=$CONF_DIR/livy_for_spark3.keytab" >> "$LIVY_CONF"
      echo "livy.server.auth.kerberos.principal=$SPNEGO_PRINCIPAL" >> "$LIVY_CONF"

      if [ "$ENABLE_THRIFT" = "true" ]; then
        echo "livy.server.thrift.authentication=kerberos" >> "$LIVY_CONF"
      fi

      # This config may contain new lines and backslashes, so it needs to be handled in a special
      # way. To preserve those characters in Java properties files, replace them with the respective
      # unicode escape sequence.
      local AUTH_TO_LOCAL=$(get_hadoop_conf "$HADOOP_CONF_DIR" "hadoop.security.auth_to_local" |
        sed 's,\\,\\u005C,g' |
        awk '{printf "%s\\u000A", $0}')
      echo "livy.server.auth.kerberos.name-rules=$AUTH_TO_LOCAL" >> "$LIVY_CONF"
    fi
  fi

  if [ "$LIVY_SESSION_RECOVERY" = "true" ]; then
    echo "livy.server.recovery.mode=recovery" >> "$LIVY_CONF"
    echo "livy.server.recovery.state-store=filesystem" >> "$LIVY_CONF"
  else
    echo "livy.server.recovery.mode=off" >> "$LIVY_CONF"
  fi

  if [ "$LIVY_RPC_ENCRYPTION" = "true" ]; then
    echo "livy.rsc.rpc.sasl.qop=auth-conf" >> "$LIVY_CONF"
  fi

  # Fix the recovery path to include the HDFS scheme.
  local DEFAULT_FS=$(get_default_fs "$HADOOP_CONF_DIR")
  local RECOVERY_DIR_KEY="livy.server.recovery.state-store.url"
  local RECOVERY_DIR=$(read_property "$RECOVERY_DIR_KEY" "$LIVY_CONF")
  RECOVERY_DIR=$(prepend_protocol "$RECOVERY_DIR" "$DEFAULT_FS")
  replace_spark_conf "$RECOVERY_DIR_KEY" "$RECOVERY_DIR" "$LIVY_CONF"

  # Hive is an optional dependency. If it's set (i.e. the configuration directory exists), then
  # enable Hive support, otherwise just create an empty directory so that Spark's scripts will
  # not try to use /etc/hive/conf.
  if [ -d "$HIVE_CONF_DIR" ]; then
    echo "livy.repl.enable-hive-context=true" >> "$LIVY_CONF"
  else
    mkdir -p "$HIVE_CONF_DIR"
  fi
}

start_livy_server() {
  export LIVY_HOME="$SPARK_HOME/livy3"
  export LIVY_CONF_DIR="$CONF_DIR/livy3-conf"
  export LIVY_SERVER_JAVA_OPTS="$CSD_JAVA_OPTS -Xmx$LIVY_HEAP_SIZE $LIVY_EXTRA_JAVA_OPTS"

  # CM retries the command a few times before giving up. If the configuration has already been
  # generated, just skip it to avoid messing things up.
  local MARKER="$CONF_DIR/start_livy_server_marker"
  if [ ! -f "$MARKER" ]; then
    gen_livy_conf
    touch "$MARKER"
  fi

  export SPARK_CONF_DIR
  export HIVE_CONF_DIR
  export SPARK_HOME

  # Point at the Spark-generated Hadoop conf, since the Spark CSD code deletes the one created
  # by CM.
  export HADOOP_CONF_DIR="$SPARK_CONF_DIR/yarn-conf"

  log "Starting Livy server..."
  exec "$LIVY_HOME/bin/livy-server"
}

gen_client_conf() {
  # Find the host/port of the Livy server and put it in the client config file.
  local SERVER_LIST="$CONF_DIR/livy3-conf/livy-server.properties"
  local CLIENT_CONF="$CONF_DIR/livy3-conf/livy-client.conf"
  local SERVER_HOST=
  local SERVER_PORT=
  local SCHEME="http"
  local THRIFT_ENABLED=
  local THRIFT_PORT=
  local THRIFT_MODE=
  for line in $(cat "$SERVER_LIST")
  do
    readconf "$line"
    case $key in
      (livy.server.port)
        SERVER_HOST="$host"
        SERVER_PORT="$value"
        ;;
      (livy.tls.enabled)
        if [ "$value" = "true" ]; then
          SCHEME="https"
        fi
        ;;
      (livy.server.thrift.enabled)
        THRIFT_ENABLED="$value"
        ;;
      (livy.server.thrift.port)
        THRIFT_PORT="$value"
        ;;
      (livy.server.thrift.transport.mode)
        THRIFT_MODE="$value"
        ;;
    esac
  done
  if [ -n "$SERVER_HOST" ]; then
    echo "livy.uri=$SCHEME://$SERVER_HOST:$SERVER_PORT" >> "$CLIENT_CONF"
  fi
  if [ "$THRIFT_ENABLED" = "true" ]; then
    echo "livy.server.thrift.enabled=$THRIFT_ENABLED" >> "$CLIENT_CONF"
    echo "livy.server.thrift.port=$THRIFT_PORT" >> "$CLIENT_CONF"
    echo "livy.server.thrift.transport.mode=$THRIFT_MODE" >> "$CLIENT_CONF"
  fi
  rm "$SERVER_LIST"
}

case $1 in
  (start_livy_server)
    start_livy_server
    ;;

  (gen_client_conf)
    gen_client_conf
    ;;

  (*)
    log "Unknown command [$1]"
    exit 1
    ;;
esac
